import logging
import re
from functools import cached_property
from re import Match

from pygls.workspace import TextDocument

from djlsp.constants import BUILTIN

logger = logging.getLogger(__name__)


class TemplateParser:
    re_loaded = re.compile(r".*{% ?load ([\w ]*) ?%}$")
    re_load = re.compile(r".*{% ?load ([\w ]*)$")
    re_url = re.compile(r""".*{% ?url ('|")([\w\-:]*)$""")
    re_static = re.compile(r".*{% ?static ('|\")([\w\-\.\/]*)$")
    re_tag = re.compile(r"^.*{% ?(\w*)$")
    re_filter = re.compile(r"^.*({%|{{) ?[\w \.\|]*\|(\w*)$")
    re_template = re.compile(r""".*{% ?(extends|include) ('|")([\w\-:]*)$""")

    def __init__(self, django_data: dict, document: TextDocument):
        self.django_data = django_data
        self.document: TextDocument = document

    @cached_property
    def loaded_libraries(self):
        loaded = {BUILTIN}
        for line in self.document.lines:
            if match := self.re_loaded.match(line):
                loaded.update(
                    [
                        lib
                        for lib in match.group(1).strip().split(" ")
                        if lib in self.django_data["libraries"]
                    ]
                )
        logger.debug(f"Loaded libraries: {loaded}")
        return loaded

    def completions(self, line, character):
        line_fragment = self.document.lines[line][:character]

        if match := self.re_load.match(line_fragment):
            return self.get_load_completions(match)
        if match := self.re_url.match(line_fragment):
            return self.get_url_completions(match)
        elif match := self.re_static.match(line_fragment):
            return self.get_static_completions(match)
        elif match := self.re_template.match(line_fragment):
            return self.get_template_completions(match)
        elif match := self.re_tag.match(line_fragment):
            return self.get_tag_completions(match)
        elif match := self.re_filter.match(line_fragment):
            return self.get_filter_completions(match)

        return []

    def get_load_completions(self, match: Match):
        prefix = match.group(1)
        logger.debug(f"Find load matches for: {prefix}")
        return sorted(
            [
                lib
                for lib in self.django_data["libraries"].keys()
                if lib != BUILTIN and lib.startswith(prefix)
            ]
        )

    def get_static_completions(self, match: Match):
        prefix = match.group(2)
        logger.debug(f"Find static matches for: {prefix}")
        return sorted(
            [
                static_file
                for static_file in self.django_data["static_files"]
                if static_file.startswith(prefix)
            ]
        )

    def get_url_completions(self, match: Match):
        prefix = match.group(2)
        logger.debug(f"Find url matches for: {prefix}")
        return sorted(
            [url for url in self.django_data["urls"] if url.startswith(prefix)]
        )

    def get_template_completions(self, match: Match):
        prefix = match.group(3)
        logger.debug(f"Find {match.group(1)} matches for: {prefix}")
        logger.debug(self.django_data["templates"])
        return sorted(
            [
                template
                for template in self.django_data["templates"]
                if template.startswith(prefix)
            ]
        )

    def get_tag_completions(self, match: Match):
        prefix = match.group(1)
        logger.debug(f"Find tag matches for: {prefix}")

        tags = []
        for name, lib in self.django_data["libraries"].items():
            if name in self.loaded_libraries:
                tags.extend(lib["tags"])

        return sorted([tag for tag in tags if tag.startswith(prefix)])

    def get_filter_completions(self, match: Match):
        prefix = match.group(2)
        logger.debug(f"Find filter matches for: {prefix}")
        filters = []
        for name, lib in self.django_data["libraries"].items():
            if name in self.loaded_libraries:
                filters.extend(lib["filters"])
        return sorted(
            [filter_name for filter_name in filters if filter_name.startswith(prefix)]
        )



import json
import os
import shutil
import subprocess

from djlsp.server import DJANGO_COLLECTOR_SCRIPT_PATH

DJANGO_TEST_PROJECT_SRC = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "django_test",
)
DJANGO_TEST_SETTINGS_MODULE = "django_test.settings"


def test_django_collect():
    django_data = json.loads(
        subprocess.check_output(
            [
                shutil.which("python"),  # Get tox python path
                DJANGO_COLLECTOR_SCRIPT_PATH,
                f"--django-settings-module={DJANGO_TEST_SETTINGS_MODULE}",
                f"--project-src={DJANGO_TEST_PROJECT_SRC}",
            ]
        )
    )

    assert "django_app" in django_data["libraries"]
    assert "django_app_tag" in django_data["libraries"]["django_app"]["tags"]
    assert "django_app_filter" in django_data["libraries"]["django_app"]["filters"]
    assert "django_app.html" in django_data["templates"]
    assert "django_app.js" in django_data["static_files"]
    assert "django_app:index" in django_data["urls"]



import argparse
import importlib
import inspect
import json
import os
import sys

import django
from django.apps import apps
from django.conf import settings
from django.contrib.staticfiles.finders import get_finders
from django.template.backends.django import get_installed_libraries
from django.template.engine import Engine
from django.template.library import InvalidTemplateLibrary
from django.template.utils import get_app_template_dirs
from django.urls import URLPattern, URLResolver

# Some tags are added with a Node, like end*, elif else.
# TODO: Find a way of collecting these, for now hardcoded list
LIBRARIES_NODE_TAGS = {
    "__builtins__": [
        # autoescape
        "endautoescape",
        # filter
        "endfilter",
        # for
        "empty",
        "endfor",
        # if
        "else",
        "elif",
        "endif"
        # ifchanged
        "endifchanged",
        # spaceless
        "endspaceless",
        # verbatim
        "endverbatim",
        # with
        "endwith",
        # block
        "endblock",
    ],
    "cache": [
        # cache    sys.path.insert(0, dance_path)
        "endcache",
    ],
    "i18n": [
        # language
        "endlanguage",
    ],
    "l10n": [
        # localize
        "endlocalize",
    ],
    "tz": [
        # localtime
        "endlocaltime",
        # timezone
        "endtimezone",
    ],
}


def get_static_files():
    # TODO: Add option to ignore some static folders
    # (like static that is generated with a JS bundler)
    static_paths = []
    for finder in get_finders():
        for path, _ in finder.list(None):
            static_paths.append(path)
    return static_paths


def get_urls():
    try:
        urlpatterns = __import__(settings.ROOT_URLCONF, {}, {}, [""]).urlpatterns
    except Exception:
        return []

    def recursive_get_views(urlpatterns, namespace=None):
        views = []
        for p in urlpatterns:
            if isinstance(p, URLPattern):
                if not p.name:
                    name = p.name
                elif namespace:
                    name = "{0}:{1}".format(namespace, p.name)
                else:
                    name = p.name
                views.append(name)
            elif isinstance(p, URLResolver):
                try:
                    patterns = p.url_patterns
                except ImportError:
                    continue
                if namespace and p.namespace:
                    _namespace = "{0}:{1}".format(namespace, p.namespace)
                else:
                    _namespace = p.namespace or namespace
                views.extend(recursive_get_views(patterns, namespace=_namespace))
        return list(filter(None, views))

    return recursive_get_views(urlpatterns)


def get_libraries():
    libraries = {
        "__builtins__": {
            "tags": [],
            "filters": [],
        }
    }

    # Collect builtins
    for lib_mod_path in Engine.get_default().builtins:
        lib = importlib.import_module(lib_mod_path).register
        libraries["__builtins__"]["tags"].extend(list(lib.tags.keys()))
        libraries["__builtins__"]["filters"].extend(list(lib.filters.keys()))

    # Get Django templatetags
    django_path = inspect.getabsfile(django.templatetags)
    django_mod_files = os.listdir(os.path.dirname(django_path))
    for django_lib in [
        i[:-3] for i in django_mod_files if i.endswith(".py") and i[0] != "_"
    ]:
        try:
            lib = get_installed_libraries()[django_lib]
            lib = importlib.import_module(lib).register
            libraries[django_lib] = {
                "tags": list(lib.tags.keys()),
                "filters": list(lib.filters.keys()),
            }
        except (InvalidTemplateLibrary, KeyError):
            continue

    for app_config in apps.get_app_configs():
        app = app_config.name
        try:
            templatetag_mod = __import__(app + ".templatetags", {}, {}, [""])
        except ImportError:
            continue

        mod_path = inspect.getabsfile(templatetag_mod)
        mod_files = os.listdir(os.path.dirname(mod_path))
        tag_files = [i[:-3] for i in mod_files if i.endswith(".py") and i[0] != "_"]

        for taglib in tag_files:
            try:
                lib = get_installed_libraries()[taglib]
                lib = importlib.import_module(lib).register
            except (InvalidTemplateLibrary, KeyError):
                continue

            libraries[taglib] = {
                "tags": list(lib.tags.keys()),
                "filters": list(lib.filters.keys()),
            }

    # Add node tags
    for lib_name, tags in LIBRARIES_NODE_TAGS.items():
        if lib_name in libraries:
            libraries[lib_name]["tags"].extend(tags)

    return libraries


def get_templates():
    template_files = []
    for templates_dir in [
        *Engine.get_default().dirs,
        *get_app_template_dirs("templates"),
    ]:
        for root, dirs, files in os.walk(templates_dir):
            for file in files:
                template_files.append(
                    os.path.relpath(os.path.join(root, file), templates_dir)
                )
    return template_files


def collect_project_data():
    return {
        "static_files": get_static_files(),
        "urls": get_urls(),
        "libraries": get_libraries(),
        "templates": get_templates(),
    }


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="""
            Collect django project information for LSP.
            pure python script that is runned in project environment for using django.
        """
    )
    parser.add_argument("--django-settings-module", action="store", type=str)
    parser.add_argument("--project-src", action="store", type=str)
    args = parser.parse_args()

    if args.django_settings_module:
        # TODO: Auto detect when empty?
        os.environ.setdefault(
            "DJANGO_SETTINGS_MODULE",
            args.django_settings_module,
        )

    if args.project_src:
        sys.path.insert(0, args.project_src)
    else:
        sys.path.insert(0, os.getcwd())

    django.setup()

    print(json.dumps(collect_project_data(), indent=4))



BUILTIN = "__builtins__"

FALLBACK_DJANGO_DATA = {
    "static_files": [],
    "urls": [],
    "templates": [],
    "libraries": {
        "__builtins__": {
            "tags": [
                "autoescape",
                "comment",
                "cycle",
                "csrf_token",
                "debug",
                "filter",
                "firstof",
                "for",
                "if",
                "ifchanged",
                "load",
                "lorem",
                "now",
                "regroup",
                "resetcycle",
                "spaceless",
                "templatetag",
                "url",
                "verbatim",
                "widthratio",
                "with",
                "block",
                "extends",
                "include",
                "endautoescape",
                "endfilter",
                "empty",
                "endfor",
                "else",
                "elif",
                "endifendifchanged",
                "endspaceless",
                "endverbatim",
                "endwith",
                "endblock",
            ],
            "filters": [
                "addslashes",
                "capfirst",
                "escapejs",
                "json_script",
                "floatformat",
                "iriencode",
                "linenumbers",
                "lower",
                "make_list",
                "slugify",
                "stringformat",
                "title",
                "truncatechars",
                "truncatechars_html",
                "truncatewords",
                "truncatewords_html",
                "upper",
                "urlencode",
                "urlize",
                "urlizetrunc",
                "wordcount",
                "wordwrap",
                "ljust",
                "rjust",
                "center",
                "cut",
                "escape",
                "force_escape",
                "linebreaks",
                "linebreaksbr",
                "safe",
                "safeseq",
                "striptags",
                "dictsort",
                "dictsortreversed",
                "first",
                "join",
                "last",
                "length",
                "length_is",
                "random",
                "slice",
                "unordered_list",
                "add",
                "get_digit",
                "date",
                "time",
                "timesince",
                "timeuntil",
                "default",
                "default_if_none",
                "divisibleby",
                "yesno",
                "filesizeformat",
                "pluralize",
                "phone2numeric",
                "pprint",
            ],
        },
        "cache": {"tags": ["cache", "endcache"], "filters": []},
        "i18n": {
            "tags": [
                "get_available_languages",
                "get_language_info",
                "get_language_info_list",
                "get_current_language",
                "get_current_language_bidi",
                "trans",
                "translate",
                "blocktrans",
                "blocktranslate",
                "language",
                "endlanguage",
            ],
            "filters": [
                "language_name",
                "language_name_translated",
                "language_name_local",
                "language_bidi",
            ],
        },
        "l10n": {
            "tags": ["localize", "endlocalize"],
            "filters": ["localize", "unlocalize"],
        },
        "static": {
            "tags": ["get_static_prefix", "get_media_prefix", "static"],
            "filters": [],
        },
        "tz": {
            "tags": [
                "localtime",
                "timezoninitializationOptionse",
                "get_current_timezone",
                "endlocaltime",
                "endtimezone",
            ],
            "filters": ["localtime", "utc", "timezone"],
        },
    },
}



import json
import logging
import os
import shutil
import subprocess

from lsprotocol.types import (
    INITIALIZE,
    TEXT_DOCUMENT_COMPLETION,
    CompletionItem,
    CompletionList,
    CompletionOptions,
    CompletionParams,
    InitializeParams,
)
from pygls.server import LanguageServer

from djlsp import __version__
from djlsp.constants import FALLBACK_DJANGO_DATA
from djlsp.parser import TemplateParser

logger = logging.getLogger(__name__)


DJANGO_COLLECTOR_SCRIPT_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "scripts",
    "django-collector.py",
)


class DjangoTemplateLanguageServer(LanguageServer):
    ENV_DIRECTORIES = [
        "env",
        ".env",
        "venv",
        ".venv",
    ]

    def __init__(self, *args):
        super().__init__(*args)
        self.docker_compose_file = "docker-compose.yml"
        self.docker_compose_service = "django"
        self.django_settings_module = ""
        self.django_data = FALLBACK_DJANGO_DATA

    def set_initialization_options(self, options: dict):
        self.docker_compose_file = options.get(
            "docker_compose_file", self.docker_compose_file
        )
        self.docker_compose_service = options.get(
            "docker_compose_service", self.docker_compose_service
        )
        self.django_settings_module = options.get(
            "django_settings_module", self.django_settings_module
        )

    def get_django_data(self):
        if self._has_valid_docker_service():
            django_data = self._get_django_data_from_docker()
        elif python_path := self._get_python_path():
            django_data = self._get_django_data_from_python_path(python_path)
        else:
            django_data = None

        if django_data:
            # TODO: Maybe validate data
            self.django_data = django_data
            logger.info("Collected project Django data:")
            logger.info(f" - Libraries: {len(django_data['libraries'])}")
            logger.info(f" - Templates: {len(django_data['templates'])}")
            logger.info(f" - Static files: {len(django_data['static_files'])}")
            logger.info(f" - Urls: {len(django_data['urls'])}")
        else:
            logger.info("Could not collect Django data")

    def _get_python_path(self):
        for env_dir in self.ENV_DIRECTORIES:
            env_python_path = os.path.join(
                self.workspace.root_path, env_dir, "bin", "python"
            )
            if os.path.exists(env_python_path):
                return env_python_path
        return shutil.which("python3")

    def _get_django_data_from_python_path(self, python_path):
        logger.info(f"Collection django data from local python path: {python_path}")
        project_src = self.workspace.root_path
        if os.path.isdir(os.path.join(project_src, "src")):
            project_src = os.path.join(project_src, "src")

        logger.info(f" - For project path: {project_src}")

        try:
            return json.loads(
                subprocess.check_output(
                    filter(
                        None,
                        [
                            python_path,
                            DJANGO_COLLECTOR_SCRIPT_PATH,
                            (
                                f"--django-settings-module={self.django_settings_module}"  # noqa: E501
                                if self.django_settings_module
                                else None
                            ),
                            f"--project-src={project_src}",
                        ],
                    )
                ).decode()
            )
        except Exception as e:
            logger.error(e)
            return False

    def _has_valid_docker_service(self):
        docker_compose_path = os.path.join(
            self.workspace.root_path, self.docker_compose_file
        )
        if os.path.exists(docker_compose_path):
            services = (
                subprocess.check_output(
                    [
                        "docker",
                        "compose",
                        f"--file={docker_compose_path}",
                        "config",
                        "--services",
                    ]
                )
                .decode()
                .splitlines()
            )
            return self.docker_compose_service in services
        return False

    def _get_django_data_from_docker(self):
        logger.info(
            f"Collecting django data from docker {self.docker_compose_file}:{self.docker_compose_service}"  # noqa: E501
        )
        docker_compose_path = os.path.join(
            self.workspace.root_path, self.docker_compose_file
        )

        try:
            return json.loads(
                subprocess.check_output(
                    filter(
                        None,
                        [
                            "docker",
                            "compose",
                            f"--file={docker_compose_path}",
                            "run",
                            "--rm",
                            f"--volume={DJANGO_COLLECTOR_SCRIPT_PATH}:/django-collector.py",  # noqa: E501
                            self.docker_compose_service,
                            "python",
                            "/django-collector.py",
                            (
                                f"--django-settings-module={self.django_settings_module}"  # noqa: E501
                                if self.django_settings_module
                                else None
                            ),
                        ],
                    )
                ).decode()
            )
        except Exception as e:
            logger.error(e)
            return False


server = DjangoTemplateLanguageServer("django-template-lsp", __version__)


@server.feature(INITIALIZE)
def initialized(ls: DjangoTemplateLanguageServer, params: InitializeParams):
    logger.info(f"COMMAND: {INITIALIZE}")
    logger.debug(f"OPTIONS: {params.initialization_options}")
    if params.initialization_options:
        ls.set_initialization_options(params.initialization_options)
    ls.get_django_data()


@server.feature(
    TEXT_DOCUMENT_COMPLETION, CompletionOptions(trigger_characters=[" ", "|", "'"])
)
def completions(ls: DjangoTemplateLanguageServer, params: CompletionParams):
    logger.info(f"COMMAND: {TEXT_DOCUMENT_COMPLETION}")
    logger.debug(f"PARAMS: {params}")
    items = []
    document = server.workspace.get_document(params.text_document.uri)
    template = TemplateParser(ls.django_data, document)
    for completion in template.completions(
        params.position.line, params.position.character
    ):
        items.append(CompletionItem(label=completion))
    return CompletionList(is_incomplete=False, items=items)
